package	 juice.example;

import 	 jam.operator.*;
import 	 jam.vector.*;
import   jam.JamObject;

/**
 * An RnOp is a real valued "matrix" which maps RnVecs to RnVecs.
 * <BR><!--#include file="./source.inc"--><BR>
 * 
 * @see         jam.RnVec
 * @see         jam.RnSpc  
 * @author      Joel Schroeder
 */
public class RnOp implements hasAdjoint {
  protected RnSpc     fwdDom, fwdRan;
  protected float[]   data;
  protected boolean   isFwd;
  protected RnOp      adj = null;

  public     RnOp(RnSpc domain, RnSpc range) {
    this.fwdDom = domain;
    this.fwdRan = range ;
    this.data   = new float[domain.size()*range.size()];
    this.isFwd  = true;
  }

  /** constructs a "square" operator (i.e. domain = range) */
  public    RnOp(RnSpc domainAndRange) {
    this(domainAndRange, domainAndRange);
  }

  /** copy constructor */
  protected RnOp(RnOp src) {
    this.fwdDom = src.fwdDom;
    this.fwdRan = src.fwdRan;
    this.isFwd  = src.isFwd ;
    this.data   = new float[src.data.length];
    for (int i = 0; i < this.data.length; i++)
      this.data[i] = src.data[i];
  }

  public JamObject copy() { return new RnOp(this); }
  
  public boolean   equals(Object obj2) {
    if (this == obj2) return true;
    if (! (obj2 instanceof RnOp)) return false;
    RnOp op2 = (RnOp) obj2;
    if (this.isFwd == op2.isFwd) {
      if (this.fwdDom != op2.fwdDom)  return false;
      if (this.fwdRan != op2.fwdRan)  return false;
    }
    else {
      if (this.fwdDom != op2.fwdRan)  return false;
      if (this.fwdRan != op2.fwdDom)  return false;
    }
    for (int i = 0; i < this.data.length; i++)
      if (this.data[i] != op2.data[i]) return false;
    return true;
  }

  public Space 	      getDomain() {
    return (isFwd) ? fwdDom:fwdRan; }
  public Space 	      getRange () {
    return (isFwd) ? fwdRan:fwdDom; }
  public Vector image    (Vector d) {
    Vector r = getRange().newMember();
    image(d, r);
    return r; }
  public void         image    (Vector d,
				Vector r) {
    ((Vector) r).zero();
    addImage(d, r);
  }
  public void         addImage (Vector d,
				Vector r) {
    int matInd = 0;
    float[] rr = ((RnVec) r).getData();
    float[] dd = ((RnVec) d).getData();
    for   (int j = 0; j < fwdDom.size(); j++ ) {
      for (int i = 0; i < fwdRan.size(); i++ ) {
	if (isFwd)
	  rr[i] += data[matInd] * dd[j];
	else
	  rr[j] += data[matInd] * dd[i];
	matInd++; }}}
  
  public void         residual (Vector d,
				Vector b,
				Vector r) {
    r.neg(b);
    addImage(d, r); }
  /**
   * Set all the matrix's components to a specified value.
   */
  public void  	      fill     (float a) {
    for (int i = 0; i < data.length; i++ ) {
      this.data[i] = a; }}
  /**
   * Set all the matrix's components to random values.
   */
  public void  	      rand     () {
    for (int i = 0; i < data.length; i++ ) {
      this.data[i] = 2*((float) Math.random()) - 1F; }}
  /**
   * Get the specified matrix entry.
   * @param rIndex The index into the range  space vector.
   * @param dIndex The index into the domain space vector.
   * @return The specified matrix entry.
   * @see RnOp#setVal
   */
  public float 	      getVal   (  int[] rIndex, int[] dIndex) {
    return data[ matrixIndex(rIndex, dIndex)]; }
  /**
   * Set the specified matrix entry.
   * 
   * @param rIndex The index into the range  space vector.
   * @param dIndex The index into the domain space vector.
   * @param a      The new value for the specified matrix entry.
   * @see RnOp#getVal
   */
  public void  	      setVal   (  int[] rIndex, int[] dIndex, float a) {
    data[ matrixIndex(rIndex, dIndex)] = a; }
  // The entries are stored with the range index being the "fast" index
  // (Is that "row major"?)
  int                 matrixIndex(int[] rIndex, int[] dIndex) {
    return (fwdRan.vecIndex(rIndex) +
	    fwdDom.vecIndex(dIndex) * fwdRan.size());
  }
  
  protected void createAdjoint() {
    adj       = new RnOp(fwdDom, fwdRan);
    adj.isFwd = !this.isFwd;
    adj.data  =  this.data;
    adj.adj   =  this;
  }

  public hasAdjoint adjoint() {
    if (adj == null) createAdjoint();
    return adj;
  }

  public LinearOperator  getGradientAt  (Vector x) {
    return adjoint();
  }
  public hasAdjoint      getDerivativeAt(Vector x) {
    return this;
  }

  public String          toString () {
    StringBuffer values = new StringBuffer();
    for (int i = 0; i < fwdDom.size() * fwdRan.size(); i++ ) {
      values.append(data[i] + " "); }
    return ("RnOp (" + (isFwd ? "forward":"adjoint") + "):" + "\n" +
	    "Domain:" + getDomain() + "\n" +
	    "Range :" + getRange () + "\n" +
	    "Values:" + values); }
}
