package juice.operator;

import    jam.operator.*;
import    jam.vector.*;
import  juice.vector.*;
import   java.util.Enumeration;

/**
 * <BR>
 * <BR><!--#include file="./source.inc"--><BR>
 *
 * @author      Joel Schroeder
 */
public class BlockOp implements BlockOperator {
  protected OpList       opList;
  protected int          nRows, nCols;
  protected Space        range, domain;

  public BlockOp(Space domain, int nCols, Space range, int nRows) {
    this.nRows  = nRows;
    this.nCols  = nCols;
    this.domain = domain;
    this.range  = range;
    this.opList = new OpList(nRows*nCols);
  }

  public BlockOp(int nRows, int nCols) {
    this((nCols > 1) ? new ProductSpc(nCols, null):null, nCols,
	 (nRows > 1) ? new ProductSpc(nRows, null):null, nRows);
  }

  /**
   * This constructor assumes that if the domain or range is a
   * <TT>ProductSpace</TT>, then the number of columns or
   * rows, respectively, equals the number of sub-spaces in the
   * product space.<BR>
   * Therefore, if the domain/range is a <TT>ProductSpace</TT>,
   * but you want the <TT>BlockOp</TT> to have only one column/row,
   * respectively, you SHOULD NOT USE THIS CONSTRUCTOR, since it
   * will make the number of columns/rows equal to the number of
   * sub-spaces in the product-space.
   */
  public BlockOp(Space domain, Space range) {
    this(domain, (domain instanceof ProductSpace) ?
	 ((ProductSpace) domain).size() : 1,
	 range , (range  instanceof ProductSpace) ?
	 ((ProductSpace) range ).size() : 1);
  }

  boolean checkOrSetSubRange (int i, Space spc) {
    Space subRange = getSubRange(i);
    if (subRange != null) {
      return subRange.equals(spc);
    }
    setSubRange(i, spc);
    return true;
  }

  boolean checkOrSetSubDomain(int j, Space spc) {
    Space subDomain = getSubDomain(j);
    if (subDomain != null) {
      return subDomain.equals(spc);
    }
    setSubDomain(j, spc);
    return true;
  }

  void setSubDomain(int j, Space spc) {
    if (nCols == 1) {
      domain = spc;
    }
    else {
      ((ProductSpace) domain).setSubSpace(j, spc);
    }
  }

  void setSubRange (int i, Space spc) {
    if (nRows == 1) {
      range  = spc;
    }
    else {
      ((ProductSpace) range ).setSubSpace(i, spc);
    }
  }

  void   setOp(int index, Operator op) {
    opList.setOp(index, op);
  }

  public void      setOp(int i, int j, Operator op) {
    if (! checkOrSetSubRange (i, op.getRange ())) {
      System.err.println("Error: range of operator passed to "     +
			 "BlockOp.setOp()\n not consistent with that of " +
			 "the BlockOp\n");
      return;
    }
    if (! checkOrSetSubDomain(j, op.getDomain())) {
      System.err.println("Error: domain of operator passed to "     +
			 "BlockOp.setOp()\n not consistent with that of " +
			 "the BlockOp\n");
      return;
    }

    setOp(index(i,j), op);
  }

  public Operator  getOp(int i, int j) {
    int ind = index(i, j);
    if (opList.getOp(ind) == null) {
      Space subRan = getSubRange (i);
      Space subDom = getSubDomain(j);
      if (subRan == null || subDom == null) {
	System.err.println("Error: BlockOp.getOp(i, j), operator not " +
			   "defined\n");
	return null;
      }
      setOp(ind, new ZeroOp(getSubRange(i), getSubDomain(j)));
    }
    return opList.getOp(ind);
  }

  Space            getSubDomain(int j) {
    if (j < 0 || j >= nCols) {
      System.err.println("Error in BlockOp.getSubDomain(): " +
			 "index out of bounds.");
      return null;
    }
    return (nCols == 1) ? domain:((ProductSpace) domain).getSubSpace(j);
  }
    
  Space            getSubRange (int i) {
    if (i < 0 || i >= nRows) {
      System.err.println("Error in BlockOp.getSubRange(): " +
			 "index out of bounds.");
      return null;
    }
    return (nRows == 1) ? range :((ProductSpace) range ).getSubSpace(i);
  }

  public int   	     nRows   () { return nRows; }
  public int   	     nCols   () { return nCols; }
  public int         size    () { return nRows * nCols; }
  public Enumeration elements() { return opList.elements(); }

  /*
  public BlockOp multiply(BlockOp blkdOp2) {
    // throw if this.getDomain() != blkdOp2.Range ()
    // The above is checked by the invoked multiply()
    BlockOp prod = new BlockOp(this   .getDomain(),
					       blkdOp2.getRange ());
    multiply(blkdOp2, prod );
    return prod;
  }

  public void            multiply(BlockOp blkdOp2,
				  BlockOp product) {
    // throw if this   .getDomain() != blkdOp2.getRange ()
    // throw if this   .getRange () != product.getRange ()
    // throw if blkdOp2.getDomain() != product.getDomain()
    int n = this.nCols();
    for     (int i = 0; i < product.nRows(); i++ ) {
      for   (int j = 0; j < product.nCols(); j++ ) {
	SummedOperator dotProduct = new SummedOperator();
	for (int k = 0; k < n              ; k++ ) {
	  dotProduct.add(new CompoundOperator(this   .getOp(i, k),
					      blkdOp2.getOp(k, j))); }
	product.setOp(i, j, dotProduct); }}
  }
  */

  public Space getDomain() { return domain; }
  public Space getRange () { return range ; }

  public Vector image   (Vector d) {
    Vector r = range.newMember();
    image(d, r);
    return r;
  }

  public void   image   (Vector d,
			 Vector r) {
    r.zero();
    addImage(d, r);
  }

  public void   addImage(Vector d,
			 Vector r) {
    for   (int i = 0; i < nRows; i++ ) {
      Vector   rr = (nRows == 1) ? r:((ProductVector) r).getSubVector(i);
      for (int j = 0; j < nCols; j++ ) {
	Vector dd = (nCols == 1) ? d:((ProductVector) d).getSubVector(j);
	getOp(i, j).addImage(dd, rr); }}
  }

  public void   residual(Vector d,
			 Vector b,
			 Vector r) {
    r.neg(b);
    addImage(d, r);
  }

  final private int index(int i, int j) { return i * nCols + j; }

  public String toString() {
    return("BlockOp, " + nRows + " by " + nCols + ":\n" +
	   "Operators:\n" + opList);
  }

  public jam.JamObject copy() {
    System.err.println("BlockOp.copy() not implemented.\n");
    return null;
  }
}
