package	 juice.operator;

import 	 jam.operator.*;
import 	 jam.vector.*;
import   jam.JamObject;

/**
 * A <TT>MatrixRep</TT> is a 2D array of floats which represent a
 * specific linear operator.  The representation is with respect to
 * two bases, one of the operator's domain, and the other of the
 * operator's range.
 *
 * @author      Joel Schroeder
 */

public class MatrixRep {
  protected LinearOperator           A;
  protected Basis          domainBasis;
  protected Basis           rangeBasis;
  protected float[][]                d;

  public MatrixRep(LinearOperator A) {
    this(A, ((SpaceBasis) A.getDomain()).getBasis(),
	 ((SpaceBasis) A.getRange()).getBasis());
  }

  public MatrixRep(LinearOperator A, Basis domainBasis, Basis rangeBasis) {
    if (! (A.getDomain().equals(domainBasis.getSpace()) &&
	   A.getRange ().equals( rangeBasis.getSpace()))) {
      System.err.println("Error in MatrixRep constructor:\n" +
			 "Space of domainBasis/rangeBasis is unequal " +
			 "to the domain/range of the operator");
      return;
    }
    this.A = A;
    this.domainBasis = domainBasis;
    this.rangeBasis  =  rangeBasis;
    calculateMatrix();
  }

  void   calculateMatrix() {
    int n = domainBasis.size();
    int m =  rangeBasis.size();
    d = new float[m][n];
    for (int i = 0; i < m; i++) {
      Vector vecI = rangeBasis.getVector(i);
      for (int j = 0; j < n; j++)
	d[i][j] = vecI.dot(A.image(domainBasis.getVector(j)));
    }
  }

  public Basis 		getDomainBasis() { return domainBasis; }
  public Basis          getRangeBasis () { return  rangeBasis; }
  public LinearOperator getOperator   () { return A; }
  public float          getValue      (int i, int j) { return d[i][j]; }

  /**
   * Converts the matrix into a form readable by Matlab.<BR>
   * <BR>Example:<BR>
   * println("A = " + myMatrixRep.toStringMatlab);<BR>
   * piping the output of this example to a file mymat.m will allow
   * you to type "mymat" at the Matlab prompt to define the
   * variable mymat to the matrix myMatrixRep.
   */
  public String toStringMatlab() {
    StringBuffer sb = new StringBuffer("[\n");
    int n = d[0].length;
    int m = d.length;
    for (int i = 0; i < m; i++) {
      for (int j = 0; j < n; j++) {
	sb.append(d[i][j] + " ");
	if (j % 40 == 39) sb.append("...\n");
      }
      sb.append("\n");
    }
    sb.append("];");
    return sb.toString();
  }
}
