package juice.operator;

import    jam.operator.*;
import    jam.vector.*;
import  juice.vector.*;
import    jam.JamObject;

/**
 * A <TT>NormResidualOpHess</TT> is a <TT>NormResidualOpGrad<TT> in which the
 * underlying operator is a <TT>hasHessian</TT> in which the adjoint
 * of the gradient has been coded!
 *
 * <BR><!--#include file="./source.inc"--><BR>
 * @author      Joel Schroeder
 */
public class NormResidualOpHess
extends NormResidualOpGrad implements hasHessian {
  public     NormResidualOpHess(hasHessian A, Vector b) { super(A, b); }
  protected  NormResidualOpHess(NormResidualOpHess src) { super(src);  }

  public boolean equals(Object obj2) {
    if (! (obj2 instanceof NormResidualOpHess)) return false;
    return super.equals(obj2);
  }

  public JamObject copy() { return new NormResidualOpHess(this); }

  public String  toString () {
    return ("NormResidualOpHess:"  + "\n" +
	    "Operator:" + this.resOp ); 
  }

  public hasAdjoint getDerivativeAt(Vector x) {
    return ((hasAdjoint) getGradientAt(x)).adjoint();
  }

  public BiLinearOperatorAdjoint getHessianAt(Vector x) {
    ResidualOp rop = (ResidualOp) resOp.getOp(0);
    return new HessianNRO((hasHessian) rop.A, rop.b, x);
  }
}

/**
 * A Hessian is symmetric, i.e. A(x,y) = A(y,x).<BR>
 * (We assume enough continuity so that partial derivatives commute.)
 */
final class HessianNRO implements BiLinearOperatorAdjoint {
  public static final int kDomain1 = 0;
  public static final int kDomain2 = 1;
  public static final int kRange   = 2;

  ProductSpace            domain;
  hasHessian              A;
  Vector                  b;
  Vector                  x;
  Vector                  res;
  BiLinearOperatorAdjoint hessianA;
  hasAdjoint              derivativeA;

  public Space getDomain1() { return domain.getSubSpace(0); }
  public Space getDomain2() { return domain.getSubSpace(1); }
  public Space getDomain () { return domain; }
  public Space getRange  () { return RealSpc.getSpace(); }

  public HessianNRO(hasHessian A, Vector b, Vector x) {
    this.A      = A;
    this.b      = b;
    this.x      = x;
    res         = b.getSpace().newMember();
    A.residual(x, b, res);
    hessianA    = A.getHessianAt(x);
    derivativeA = ((hasAdjoint) A.getGradientAt(x)).adjoint();
  }

  /** copy constructor */
  HessianNRO(HessianNRO src) {
    this((hasHessian) src.A.copy(), (Vector) src.b.copy(),
	 (Vector) src.x.copy());
  }

  public Vector       image(Vector d) {
    if (! (d instanceof ProductVector)) {
      System.err.println("Error in HessianNRO.image(): " +
			 "Vector argument not a ProductVector");
      return null;
    }
    return imageBinary(((ProductVector) d).getSubVector(0),
		       ((ProductVector) d).getSubVector(1));
  }

  public void         image(Vector d, Vector r) {
    if (! (d instanceof ProductVector)) {
      System.err.println("Error in HessianNRO.image(): " +
			 "Vector argument not a ProductVector");
      return;
    }
    imageBinary(((ProductVector) d).getSubVector(0),
		((ProductVector) d).getSubVector(1), r);
  }

  public void      addImage(Vector d, Vector r) {
    if (! (d instanceof ProductVector)) {
      System.err.println("Error in HessianNRO.addImage(): " +
			 "Vector argument not a ProductVector");
      return;
    }
    addImageBinary(((ProductVector) d).getSubVector(0),
		   ((ProductVector) d).getSubVector(1), r);

  }

  public void      residual(Vector d, Vector b, Vector r) {
    if (! (d instanceof ProductVector)) {
      System.err.println("Error in HessianNRO.residual(): " +
			 "Vector argument not a ProductVector");
      return;
    }
    residualBinary(((ProductVector) d).getSubVector(0),
		   ((ProductVector) d).getSubVector(1), b, r);

  }

  public Vector      imageBinary(Vector d1, Vector d2          ) {
    Vector r = RealSpc.getSpace().newMember();
    apply(false, d1, d2, r, kRange);
    return r;
  }
  public void        imageBinary(Vector d1, Vector d2, Vector r) {
    apply(false, d1, d2, r, kRange);
  }
  public void     addImageBinary(Vector d1, Vector d2, Vector r) {
    apply(true , d1, d2, r, kRange);
  }
  public void     residualBinary(Vector d1, Vector d2,
 			         Vector b , Vector r ) {
    r.neg(b);
    apply(true , d1, d2, r, kRange);
  }

  public void apply(boolean add,
		    Vector d1, Vector d2, Vector r,
		    int result) {
    if (result == kRange) {
      float rr = res.dot(hessianA.imageBinary(d1, d2));
      rr      += derivativeA.image(d1).dot(derivativeA.image(d2));
      if (add)
	rr += ((RealVector) r).getValue();
      ((RealVector) r).setValue(2*rr);
      return;
    }
    if (result == kDomain2) { // swap d1 and d2 since the Hessian is symmetric
      Vector tmp = d1;
      d1 = d2;
      d2 = tmp;
      result = kDomain1;
    }
    if (result == kDomain1) {
      if (!add) d1.zero();
      Vector dd = hessianA.getPartialAdjoint(res).image(d2);
      dd.add(derivativeA.adjoint().image(derivativeA.image(d2)));
      d1.addScale(2 * ((RealVector) r).getValue(), dd);
      return;
    }
    System.err.println("Error: integer argument \"result\" is invalid:\n" +
		       "       result = " + result);
  }

  public Operator getOp1(Vector d1) {
    return new HessianSubOpNRO(this, d1);
  }
  public hasAdjoint getPartialAdjoint(Vector r) {
    return new HessianAdjNRO(this, r);
  }

  public Operator getOp2(Vector d2) { return getOp1(d2); }
  public JamObject copy() {
    return new HessianNRO(this);
  }
  public boolean equals(Object obj2) {
    if (this == obj2) return true;
    if (! (obj2 instanceof HessianNRO)) return false;
    HessianNRO hes2 = (HessianNRO) obj2;
    return (this.A.equals(hes2.A) &&
	    this.b.equals(hes2.b) &&
	    this.x.equals(hes2.x));
  }
}

final class HessianSubOpNRO implements hasAdjoint {
  HessianNRO      hessian;
  HessianSubOpNRO adj   = null;
  boolean    	  isFwd = true;
  Vector     	  d0;

  public HessianSubOpNRO(HessianNRO hessian, Vector dKnown) {
    this.hessian = hessian;
    d0 = dKnown;
  }

  public HessianSubOpNRO(HessianSubOpNRO src) {
    this((HessianNRO) src.hessian.copy(), (Vector) src.d0.copy());
    this.isFwd   = src.isFwd;
  }

  public Space getDomain() {
    return isFwd ? getForwardDomain():hessian.getRange();
  }
  public Space getRange () {
    return isFwd ? hessian.getRange():getForwardDomain();
  }
  Space        getForwardDomain()  {
    return ((ProductSpace) hessian.getDomain()).getSubSpace(0);
  }

  public Vector       image(Vector d) {
    return hessian.imageBinary(d0, d);
  }
  public void         image(Vector d, Vector r) {
    hessian.imageBinary(d0, d, r);
  }
  public void      addImage(Vector d, Vector r) {
    hessian.addImageBinary(d0, d, r);
  }
  public void      residual(Vector d, Vector b, Vector r) {
    hessian.residualBinary(d0, d, b, r);
  }

  void createAdjoint() {
    adj = new HessianSubOpNRO(this.hessian, this.d0);
    adj.isFwd = !this.isFwd;
    adj.adj   = this;
  }

  public hasAdjoint     adjoint() {
    if (adj == null) createAdjoint();
    return adj;
  }
  public LinearOperator getGradientAt  (Vector x) { return adjoint(); }
  public hasAdjoint     getDerivativeAt(Vector x) { return this;      }

  public JamObject copy() { return new HessianSubOpNRO(this); }
  public boolean equals(Object obj2) {
    if (this == obj2) return true;
    if (! (obj2 instanceof HessianSubOpNRO)) return false;
    HessianSubOpNRO op2 = (HessianSubOpNRO) obj2;
    return (this.hessian.equals(op2.hessian) &&
	    this.isFwd == op2.isFwd);
  }
}

/** the partial adjoint is self-adjoint since the Hessian is symmetric */
final class HessianAdjNRO implements hasAdjoint {
  HessianNRO hessian;
  Vector     r0;

  public HessianAdjNRO(HessianNRO hessian, Vector rKnown) {
    this.hessian = hessian;
    r0 = rKnown;
  }

  public HessianAdjNRO(HessianAdjNRO src) {
    this((HessianNRO) src.hessian.copy(), (Vector) src.r0.copy());
  }

  public Space getRange () { return getDomain(); }
  public Space getDomain() {
    return ((ProductSpace) hessian.getDomain()).getSubSpace(0);
  }

  public Vector       image(Vector d) {
    Vector r = getRange().newMember();
    hessian.apply(false, d, r, r0, HessianNRO.kDomain2);
    return r;
  }
  public void         image(Vector d, Vector r) {
    hessian.apply(false, d, r, r0, HessianNRO.kDomain2);
  }
  public void      addImage(Vector d, Vector r) {
    hessian.apply(true , d, r, r0, HessianNRO.kDomain2);
  }
  public void      residual(Vector d, Vector b, Vector r) {
    r.neg(b);
    hessian.apply(true , d, r, r0, HessianNRO.kDomain2);
  }

  public hasAdjoint     adjoint()                 { return this; }
  public LinearOperator getGradientAt  (Vector x) { return this; }
  public hasAdjoint     getDerivativeAt(Vector x) { return this; }

  public JamObject copy() { return new HessianAdjNRO(this); }
  public boolean equals(Object obj2) {
    if (this == obj2) return true;
    if (! (obj2 instanceof HessianAdjNRO)) return false;
    HessianAdjNRO op2 = (HessianAdjNRO) obj2;
    return (this.hessian.equals(op2.hessian));
  }
}
