package juice.solver;

import    jam.solver.*;
import    jam.vector.*;
import    jam.operator.*;
import  juice.vector.*;
import  juice.operator.*;

/**
 * Conjugate Gradient Solver.<BR>
 * <BR><!--#include file="./source.inc"--><BR>
 *
 * Maybe this thing needs a factory? 
 *
 * @see       CGNonlinPrecondSolver
 * @see       CGSolver
 * @author    Matt Schwab
 * @version   22 Mar 1997 
 */
public class CGNonlinSolver extends IterativeSlv implements Solver {

  // the next could be in IterativeSlv
  protected Space  xSpace;
  protected Space  bSpace; 
  protected DirectionScout     directionScout;
  protected ApproximationScout approximationScout;

  protected boolean hasMoreIterations = true; 
  protected Vector  r;      //residual
  protected Vector  p;      //direction vector
  protected float   rtrNew; //rtr Norm of the residual

  // constructors 
  /**
   * @param xSpace is the model space
   * @param bSpace is the data  space
   */
  public CGNonlinSolver(Space xSpace, Space bSpace) {
    this(xSpace, bSpace,
	 new SecantLineScout(xSpace), 
	 new FletcherReeves(), 
	 new IteratorSimple()); 
  }
  public CGNonlinSolver(Space xSpace, Space bSpace, Iterator iterator) {
    this(xSpace, bSpace,
	 new SecantLineScout(xSpace), 
	 new FletcherReeves(), 
	 iterator); 
  }

  /**
   * @param xSpace is the model space
   * @param bSpace is the data  space
   * @param approximationScout is updates solution x at each iteration
   * @param directionScout is updates the search direction at each iteration
   * @param iterator marshals the iteration process
   */
  public CGNonlinSolver(Space xSpace, Space bSpace, 
			ApproximationScout approximationScout,
			DirectionScout     directionScout, 
			Iterator           iterator) {
    this.xSpace             = xSpace;
    this.bSpace             = bSpace;
    this.directionScout     = directionScout;
    this.approximationScout = approximationScout;
    this.iterator = iterator;
    this.iterator.setSolver(this); 
  }

  // communicator (mostly to Iterator). Defaults are in IterativeSlv.

  public Vector getResidual() {
    return r;
  }
  public float  getNormResidual() {
    return rtrNew;
  }

  // helper objects 

  /**
   * chooses FletcherReeves' beta for finding the next direction.
   */
  public void setFletcherReevesScout() { 
    directionScout = new FletcherReeves();
  }
  /**
   * @param resetMax sets the maximal number of iterations before 
   *        ensuring conjugacy (Tip: try resetMax = the number of unknowns).
   */
  public void setFletcherReevesScout(int resetMax) {
    directionScout = new FletcherReeves(resetMax);
  } 
  /**
   * chooses Polak-Ribiere's beta for finding the next direction.
   */    
  public void setPolakRibiereScout()  { 
    directionScout = new PolakRibiere(xSpace);
  }
  /**
   * chooses Polak-Ribiere's beta for finding the next direction.
   *
   * @param xSpace the model space 
   * @param resetMax sets the maximal number of iterations before 
   *        ensuring conjugacy (Tip: try resetMax = the number of unknowns).
   */    
  public void setPolakRibiereScout(int resetMax) {
    directionScout = new PolakRibiere(resetMax, xSpace);
  }

  /**
   * chooses the secant method to do the line search for the x approximation
   * (alpha) step.<BR>
   * @param Sigma is a starting step-length. It is dynamically readjusted
   * @param alphaMin is the minimum value of alpha the search finds acceptable
   * @param maxSteps is the maximum number of steps the search can take
   *        (a small maxSteps (inexact line search) is often sufficient)
   * @param xSpace is the space of the approximation x<BR> 
   */
  public void setSecantLineScout(float sigma, float alphaMin, int maxSteps) {
    approximationScout = new SecantLineScout(sigma, alphaMin, maxSteps, xSpace);
  }
  /**
   * chooses the secant method to do the line search for the x approximation
   * (alpha) step.<BR>
   */
  public void setSecantLineScout() {
    approximationScout = new SecantLineScout(xSpace);
  }

  /**
   * chooses the Newton-Raphson method to do the line search for 
   * the x approximation (alpha) step.<BR>
   * @param alphaMin is the minimum value of alpha the search finds acceptable
   * @param maxSteps is the maximum number of steps the search can take
   */
  public void setNewtonRaphsonScout(float alphaMin, int maxSteps) {
    approximationScout = new NewtonRaphsonScout(alphaMin, maxSteps);
  }
  public void setNewtonRaphsonScout() {
    approximationScout = new NewtonRaphsonScout();
  }

  // solve mechanism

  public void solve(Operator A, Vector b, Vector x) {
    System.err.println("Error to CGNonlinSolver.solve(A, b, x):\n" +
		       "A must be hasGradient or hasHessian.");
  }

  public void solve(hasGradient A, Vector b, Vector x) { 
    NormResidualOpGrad f = new NormResidualOpGrad((hasGradient) A, b);
    this.solve(f, x);
  }

  public void solve(hasHessian A, Vector b, Vector x) { 
    NormResidualOpHess f = new NormResidualOpHess((hasHessian)  A, b);
    this.solve(f, x);
  }

  public void solve(hasGradient f, Vector x) { 
    if (! (f.getRange() instanceof RealSpace)) {
      System.err.println("Error in CGNonlinSolver.solve(): " +
			 "operator f must be a functional");
    }
    if (approximationScout == null) {
      System.err.println("Warning: CGNonlinSolver has no " +
			 "approximationScout.\n  Default is SecantLineScout.");
      setSecantLineScout();
    }
    if (    directionScout == null) {
      System.err.println("Warning: CGNonlinSolver has no " +
			 "directionScout.\n  Default is FletcherReevesScout.");
      setFletcherReevesScout();
    }
    if (iterator == null) {
      System.err.println("Warning: CGNonlinSolver has no " +
			 "iterator.\n  Default is IteratorSimple.");
      iterator = new IteratorSimple();
    }

    this.A = f;
    this.x = x;
    this.b = new RealVec(0f); // included for completeness

    this.start();
    iterator.start();

    while(iterator.hasMoreIterations() && this.hasMoreIterations()) {
      this.nextIteration();
      iterator.nextIteration();
      iter++;
    }
    this.finish();
    iterator.finish();
  }

  /**
   * initializes before the first iteration
   */
  protected void start() {
    r = ((hasGradient) A).getGradientAt(x).image(new RealVec(-1F)); // r = - gradient
    rtrNew  = r.norm2();
    p = (Vector) r.copy();
    approximationScout.start(A,x,p);      //is not doing anything currently
    directionScout.start(r, rtrNew, p);
  } 

  /**
   * executed at each iteration
   */
  protected void nextIteration(){
    //updates x;
    approximationScout.update(A, x, p, hasMoreIterations);
    //System.out.println("   Here is x:" + x);

    //updates r
    r = ((hasGradient) A).getGradientAt(x).image(new RealVec(1F));
    r.neg();
    rtrNew = r.norm2();
    //System.out.println("   Here is r:" + r);

    //updates p;
    directionScout.update(r, rtrNew, p, hasMoreIterations);
    //System.out.println("   Here is p:" + p);

  }

  /**
   * executed after the last iteration
   */
  protected void finish() {} 
  protected boolean hasMoreIterations() {
    return hasMoreIterations;
  } 
}

/**
 * objects that implement this interface update the approximation x 
 * during an iterative conjugate gradient optimization.<BR>
 *
 * @see CGNonlinSolver
 * @see SecantLineSolver
 * @see DirectionScout
 */
abstract class ApproximationScout {
  /** 
   * initializes the object before the first solver iteration 
   * @param A is an operator (possibly gradient or hessian required)
   * @param x is the approximation 
   * @param p is the search direction
   */
  protected abstract void start (Operator A, Vector x, Vector p);
  /** 
   * updates the approximation at each solver iteration 
   * @param f is the operator (hasGradient)
   * @param x is the approximation 
   * @param p is the search direction
   * @param hasMoreIterations is set to false if this object 
   *        decides that the solver iterations are to be terminated.
   */
  protected abstract void update(Operator A, Vector x, Vector p, 
				 boolean hasMoreIterations);
}

/**
 * this updates the approximation x. It uses a Secant method to find 
 * the optimal stepthlength alpha.  
 * 
 * @see ApproximationScout
 * @author Matt Schwab
 */ 
class SecantLineScout extends ApproximationScout {
  float  alpha;
  float  alphaMin;
  int    maxSteps;
  float  secantSigma;
  float  ptp; 
  float  etaOld, etaNew; 
  Vector tmpVector;
  /**
   * @param xSpace is the space of the approximation x<BR> 
   */
  protected SecantLineScout(Space xSpace) {
    this(0.1f, 0.00001f, 10, xSpace);
  }
  /**
   * @param Sigma is a starting step-length. It is dynamically readjusted
   * @param alphaMin is the minimum value of alpha the search finds acceptable
   * @param maxSteps is the maximum number of steps the search can take
   * @param xSpace is the space of the approximation x
   */
  protected SecantLineScout(float secantSigma, 
			    float alphaMin, 
			    int   maxSteps,
			    Space xSpace) {
    tmpVector        = xSpace.newMember();
    this.secantSigma = secantSigma;
    this.alphaMin    = alphaMin*alphaMin;
    this.maxSteps    = maxSteps;
  }
  protected void start (Operator f, Vector x, Vector p) {}
  protected void update(Operator f, Vector x, Vector p, 
			boolean hasMoreIterations){
    ptp                 =   p.norm2();
    alpha               = -secantSigma; 
    tmpVector.addScale(secantSigma, x, p);
    float  etaOld       = ((hasGradient) f).getGradientAt(tmpVector)
                                           .image(new RealVec(1F)).dot(p);
    int alphaCounter    =   0;
    do {
      float etaNew      = ((hasGradient) f).getGradientAt(x)
                                           .image(new RealVec(1F)).dot(p);
      alpha            *= etaNew / (etaOld - etaNew);
      //System.out.println("alpha is : " + alpha);
      x.addScale(alpha, x, p);
      etaOld            = etaNew;
      alphaCounter++;
    } while (alphaCounter < maxSteps && alpha*alpha*ptp > alphaMin); 
  }
}

/**
 * this updates the approximation x. It uses a Newton-Raphson method to find 
 * the optimal step length alpha. <BR>
 * 
 * FUTURE STUDY:<BR>
 * What are the factors which influence a good choice of alphaMin and maxSteps?
 * 
 * @see ApproximationScout
 * @author Matt Schwab
 */ 
class NewtonRaphsonScout extends ApproximationScout {
  float   alpha;
  float   alphaMin;
  int     alphaCounter;
  int     maxSteps;
  RealVec tmpVector;
  float   ptp; 
  float   eta1, eta2; 

  protected NewtonRaphsonScout() { this(0.00001f, 10); }
  /**
   * @param alphaMin is the minimum value of alpha the search finds acceptable
   * @param maxSteps is the maximum number of steps the search can take
   */
  protected NewtonRaphsonScout(float alphaMin, int maxSteps) {
    tmpVector        = new RealVec();
    this.alphaMin    = alphaMin*alphaMin;
    this.maxSteps    = maxSteps;
  }
  protected void start (Operator f, Vector x, Vector p) {}
  protected void update(Operator f, Vector x, Vector p, 
			boolean hasMoreIterations){
    ptp          =   p.norm2();
    alphaCounter =   0;
    do {
      tmpVector = (RealVec) ((hasHessian) f).getHessianAt(x).imageBinary(p, p);
      eta2  = tmpVector.getValue();
      eta1  = ((hasGradient) f).getGradientAt(x).image(new RealVec(1F)).dot(p);
      alpha = - eta1/eta2;
      //System.out.println("alpha is : " + alpha);
      x.addScale(alpha, x, p);
      alphaCounter++;
    } while (alphaCounter < maxSteps && alpha*alpha*ptp > alphaMin); 
  }
}

/**
 * objects that implement this interface update the direction p 
 * during an iterative conjugate gradient optimization.<BR>
 *
 * @see CGNonlinSolver
 * @see FletcherReeves
 * @see ApproximationScout
 */
abstract class DirectionScout { 
  protected int resetCounter = 0;
  protected int resetMax; 
  protected float beta; 
  /** 
   * initializes the object before the first solver iteration 
   * @param r is the residual 
   * @param rtr is the norm of the residual
   * @param p is the search direction
   */
  abstract protected void start (Vector r, float rtr, Vector p);
  /** 
   * initializes the object before the first solver iteration 
   * @param r is the residual 
   * @param rtr is the norm of the residual
   * @param p is the search direction
   * @param hasMoreIterations is set to false if this object 
   *        decides that the solver iterations are to be terminated.
   */
  abstract protected void update(Vector r, float rtr, Vector p, 
				 boolean hasMoreIterations);
}

/**
 * this always converges. If the starting value is sufficiently close 
 * to the desired minimum, you are guaranteed to converge to that minimum.
 *
 * After each reset-th iteration I ensure conjugacy.<BR>
 * This is important in nonlinear CG solvers since the A-conjugacy 
 * (conjugacy with regard to the operator) deterioates as the 
 * operator itself changes.<BR>
 * I  wish I had a good idea on how to guess a good starting value. 
 * resetMax may want to be the dimension of x: but currently our 
 * abstract vector space does not include a size method. Rsf does.
 * I default it to 50 for the time being.  
 */
class FletcherReeves extends DirectionScout {
  float rtrOld; 
  FletcherReeves() { this(10); }
  FletcherReeves(int resetMax) {
    this.resetMax = resetMax;
  }
  protected void start (Vector r, float rtrNew, Vector p) {
    this.rtrOld = rtrNew;
  }
  protected void update(Vector r, float rtrNew, Vector p, 
			boolean hasMoreIterations) {
    beta   = rtrNew/rtrOld;
    //System.out.println("beta is: " + beta); 
    rtrOld = rtrNew;
    p.addScale(beta, r, p);

    resetCounter++;
    if(resetCounter >= resetMax || r.dot(p) <= 0f || beta <= 0f) {
System.out.println("I reset ..... " + beta);
      p.zero();
      p.add(r);
      resetCounter = 0;
    }    
  } 
}

/**
 * this converges faster than Fletcher-Reeves method, but 
 * it can cycle infinitely in some rare cases (I suggest you 
 * limit the number of iterations). Additionally, it stores 
 * a temporary residual vector. 
 */
class PolakRibiere extends DirectionScout {
  Vector tmpVector; // stores the previous residual
  float rtrOld;
  float rtm;
  PolakRibiere(Space xSpace) { this(10, xSpace); }
  PolakRibiere(int resetMax, Space xSpace) {
    this.resetMax = resetMax;
  }
  protected void start (Vector r, float rtrNew, Vector p) {
    this.rtrOld = rtrNew;
    tmpVector = (Vector) r.copy();
  }
  protected void update(Vector r, float rtrNew, Vector p, 
			boolean hasMoreIterations) {
    rtm = tmpVector.dot(r);
    beta = (rtrNew - rtm) / rtrOld;
//System.out.println("beta is: " + beta); 
    rtrOld = rtrNew;

    resetCounter++;
    if(resetCounter >= resetMax || beta <= 0f) {
System.out.println("I reset ..... " + beta);
      p.zero();
      p.add(r);
      resetCounter = 0;
    }
    else {
      p.addScale(beta, r, p); 
    }
    tmpVector.zero();
    tmpVector.add(r);
  }
}




