package juice.solver;

import  juice.example.*;
import  juice.vector.*;
import  juice.operator.*;
import  juice.util.Tester;
import    jam.solver.*;
import    jam.vector.*;
import    jam.operator.*;
import    jam.JamObject;      
import   java.util.Random;
/**
 * <BR><!--#include file="./source.inc"--><BR>
 */
final
public class CGNonlinSolverTest {

  /**
   * I solve here 0 = F.image(x) for x = (x1, x2)
   * where F is the Rosenbrook's function 
   * F(x1,x2) = 100(x2-x1^2)^2 + (1- x1)^2
   * The unique minimum is x = (1,1). I use x = (-1.2, 1.0) 
   * as a starting value. 
   */ 
  public static void main( String[] args) {
    System.out.println( "CGNonlinSolver Test\n------------------\n");

    RnVec      x = new RnVec(2);
    int[] index0 = { 0 }; x.setVal(index0, -1.2f);
    int[] index1 = { 1 }; x.setVal(index1,  1.0f);

    RnVec      m = new RnVec(2);
    m.setVal(index0,  1.0f);
    m.setVal(index1,  1.0f);

    RealVec b = new RealVec(0f);

    RosenbrookOp F = new RosenbrookOp();
    LinearOperator fgrad = F.getGradientAt(x);
System.out.println("Here is fgrad(x): " + fgrad.image(new RealVec(1f)));

System.out.println("Here is F(x): " + F.image(x));
System.out.println("Here is F(m): " + F.image(m));

    CGNonlinSolver mysolver = new CGNonlinSolver(F.getDomain(), F.getRange());
    mysolver.setPolakRibiereScout();
    mysolver.setNewtonRaphsonScout();
    mysolver.solve(F, x); 

System.out.println("Here is x: " + x);
System.out.println("Here is m: " + m);

    // e = |x - m| 
    RnVec    e = new RnVec((RnSpc) x.getSpace());
    e.sub(x, m);
    System.out.println("|X-x| = " + e.norm());
    Tester.exit(e.norm() < 10e-5);
  }
}

class RosenbrookOp implements hasHessian {
  RnSpc   domain;

  public RosenbrookOp()    { domain = new RnSpc(2); }
  public Space getDomain() { return domain; }
  public Space getRange()  { return RealSpc.getSpace(); }
  public jam.JamObject copy() { 
    System.err.println("RosenbrookOp: copy() is simply a clone"); 
    try {
      return (jam.JamObject) this.clone();
    }
    catch (Exception e) { System.err.println(e); }
    return null;
  }

  public void   image(Vector x, Vector d) {
    int[] index0 = {0}; float x1 = ((RnVec) x).getVal(index0); 
    int[] index1 = {1}; float x2 = ((RnVec) x).getVal(index1); 
    float tmp = 100f*(x2 - x1 * x1)*(x2 - x1*x1) + (1 - x1)*(1 - x1);
    ((RealVec) d).setValue(tmp); 
  }

  public Vector image(Vector x) {
    RealVec d = new RealVec();
    this.image(x, (Vector) d);
    return d;
  }
  public void addImage(Vector x, Vector d){ 
    d.add(this.image(x));
  }
  public void residual(Vector d, Vector b, Vector r) {
    r.sub(this.image(d),b);
  }
  public LinearOperator getGradientAt(Vector x) { 
    return new RosenbrookGrad(x, (RnSpc) getDomain());
  }
  public hasAdjoint     getDerivativeAt(Vector x) {
    return ((hasAdjoint) getGradientAt(x)).adjoint();
  }
  public BiLinearOperatorAdjoint getHessianAt(Vector x) {
    return new RosenbrookHess(x);
  }
}

class RosenbrookGrad implements hasAdjoint {
  RnSpc      	 fwdRange;
  Vector     	 x;
  RosenbrookGrad adj   = null;
  boolean        isFwd = true;

  public RosenbrookGrad(Vector x, RnSpc range) {
    this.fwdRange  = range;
    this.x         = x;
  }

  RosenbrookGrad(RosenbrookGrad src) {
    this((Vector) src.x.copy(), (RnSpc) src.fwdRange.copy());
    this.isFwd = src.isFwd;
  }

  public Space getDomain() {
    if (isFwd) return RealSpc.getSpace();
    else       return fwdRange;
  }
  public Space getRange () {
    if (isFwd) return fwdRange;
    else       return RealSpc.getSpace();
  }
  public void image(Vector d, Vector r) {
    int[] index0 = {0}; float x1 = ((RnVec) x).getVal(index0); 
    int[] index1 = {1}; float x2 = ((RnVec) x).getVal(index1); 
    float f1 = -400f*x1*(x2 - x1 * x1) - 2*(1 - x1);
    float f2 =  200f*   (x2 - x1 * x1);
    if (isFwd) {
      float real = ((RealVec) d).getValue();
      RnVec rn   = (RnVec)   r;
      rn.setVal(index0, real * f1); 
      rn.setVal(index1, real * f2); 
    }
    else {
      RnVec   rn   = (RnVec)   d;
      RealVec real = (RealVec) r;
      real.setValue(rn.getVal(index0) * f1 + rn.getVal(index1) * f2);
    }
  }    
  public Vector image(Vector d) {
    Vector r = getRange().newMember();
    this.image(d, r);
    return r;
  }
  public void addImage(Vector d, Vector r){ 
    r.add(this.image(r));
  }
  public void residual(Vector d, Vector b, Vector r) {
    r.sub(this.image(d),b);
  }

  public jam.JamObject copy() { return new RosenbrookGrad(this); }

  void createAdjoint() {
    adj = new RosenbrookGrad(x, fwdRange);
    adj.isFwd = !this.isFwd;
    adj.adj = this;
  }

  public hasAdjoint adjoint() {
    if (adj == null) createAdjoint();
    return adj;
  }

  public LinearOperator getGradientAt  (Vector x) { return adjoint(); }
  public hasAdjoint     getDerivativeAt(Vector x) { return this; }
}

class RosenbrookHess implements BiLinearOperatorAdjoint {
  public static final int kDomain1 = 0;
  public static final int kDomain2 = 1;
  public static final int kRange   = 2;

  ProductSpace domain;
  float[][]    f;

  public RosenbrookHess(Vector x) {
    domain = new ProductSpc(x.getSpace(), x.getSpace());
    createMatrix((RnVec) x);
  }

  /** copy constructor */
  RosenbrookHess(RosenbrookHess src) {
    this.f = new float[2][2];
    this.f[0][0] = src.f[0][0];
    this.f[0][1] = src.f[0][1];
    this.f[1][0] = src.f[1][0];
    this.f[1][1] = src.f[1][1];
    this.domain = (ProductSpace) src.domain.copy();
  }

  void createMatrix(RnVec x) {
    int[] index0 = {0}; float x1 = x.getVal(index0); 
    int[] index1 = {1}; float x2 = x.getVal(index1); 
    f  = new float[2][2];
    f[0][0] = -400*x2 + 1200*x1*x1 + 2; f[0][1] = -400*x1;
    f[1][0] = f[0][1];                  f[1][1] = 200;
  }

  public Space getDomain1() { return domain.getSubSpace(0); }
  public Space getDomain2() { return domain.getSubSpace(1); }
  public Space getDomain () { return domain; }
  public Space getRange  () { return RealSpc.getSpace(); }

  public Vector       image(Vector d) {
    if (! (d instanceof ProductVector)) {
      System.err.println("Error in RosenbrookHess.image(): " +
			 "Vector argument not a ProductVector");
      return null;
    }
    return imageBinary(((ProductVector) d).getSubVector(0),
		       ((ProductVector) d).getSubVector(1));
  }

  public void         image(Vector d, Vector r) {
    if (! (d instanceof ProductVector)) {
      System.err.println("Error in RosenbrookHess.image(): " +
			 "Vector argument not a ProductVector");
      return;
    }
    imageBinary(((ProductVector) d).getSubVector(0),
		((ProductVector) d).getSubVector(1), r);
  }

  public void      addImage(Vector d, Vector r) {
    if (! (d instanceof ProductVector)) {
      System.err.println("Error in RosenbrookHess.addImage(): " +
			 "Vector argument not a ProductVector");
      return;
    }
    addImageBinary(((ProductVector) d).getSubVector(0),
		   ((ProductVector) d).getSubVector(1), r);

  }

  public void      residual(Vector d, Vector b, Vector r) {
    if (! (d instanceof ProductVector)) {
      System.err.println("Error in RosenbrookHess.residual(): " +
			 "Vector argument not a ProductVector");
      return;
    }
    residualBinary(((ProductVector) d).getSubVector(0),
		   ((ProductVector) d).getSubVector(1), b, r);

  }

  public Vector      imageBinary(Vector d1, Vector d2          ) {
    Vector r = RealSpc.getSpace().newMember();
    apply(false, d1, d2, r, kRange);
    return r;
  }
  public void        imageBinary(Vector d1, Vector d2, Vector r) {
    apply(false, d1, d2, r, kRange);
  }
  public void     addImageBinary(Vector d1, Vector d2, Vector r) {
    apply(true , d1, d2, r, kRange);
  }
  public void     residualBinary(Vector d1, Vector d2,
 			         Vector b , Vector r ) {
    r.neg(b);
    apply(true , d1, d2, r, kRange);
  }

  public void apply(boolean add,
		    Vector d1, Vector d2, Vector r,
		    int result) {
    int[] index0 = {0};
    int[] index1 = {1};
    if (result == kRange) {
      float rr  = ((RnVec) d1).getVal(index0) * ((RnVec) d2).getVal(index0)
	* f[0][0];
      rr       += ((RnVec) d1).getVal(index0) * ((RnVec) d2).getVal(index1)
	* f[0][1];
      rr       += ((RnVec) d1).getVal(index1) * ((RnVec) d2).getVal(index0)
	* f[1][0];
      rr       += ((RnVec) d1).getVal(index1) * ((RnVec) d2).getVal(index1)
	* f[1][1];
      if (add)
	rr += ((RealVector) r).getValue();
      ((RealVector) r).setValue(rr);
      return;
    }
    if (result == kDomain2) { // swap d1 and d2 since the Hessian is symmetric
      Vector tmp = d1;
      d1 = d2;
      d2 = tmp;
      result = kDomain1;
    }
    if (result == kDomain1) {
      float d20 = ((RnVec) d2).getVal(index0);
      float d21 = ((RnVec) d2).getVal(index1);
      float x0 = f[0][0] * d20 + f[0][1] * d21;
      float x1 = f[1][0] * d20 + f[1][1] * d21;
      if (add) {
	x0 += ((RnVec) d1).getVal(index0);
	x1 += ((RnVec) d1).getVal(index1);
      }	
      ((RnVec) d1).setVal(index0, x0);
      ((RnVec) d1).setVal(index1, x1);
      return;
    }
    System.err.println("Error: integer argument \"result\" is invalid:\n" +
		       "       result = " + result);
  }

  public Operator getOp1(Vector d1) {
    return new RHessSubOp(this, d1);
  }
  public hasAdjoint getPartialAdjoint(Vector r) {
    return new RHessAdj(this, r);
  }

  public Operator getOp2(Vector d2) { return getOp1(d2); }
  public JamObject copy() { return new RosenbrookHess(this); }
  public boolean equals(Object obj2) {
    if (this == obj2) return true;
    if (! (obj2 instanceof RosenbrookHess)) return false;
    System.err.println("Equals not well implemented for RosenbrookHess\n");
    return false;
  }
}

final class RHessSubOp implements hasAdjoint {
  RosenbrookHess hessian;
  RHessSubOp     adj   = null;
  boolean    	 isFwd = true;
  Vector     	 d0;

  public RHessSubOp(RosenbrookHess hessian, Vector dKnown) {
    this.hessian = hessian;
    d0 = dKnown;
  }

  public RHessSubOp(RHessSubOp src) {
    this((RosenbrookHess) src.hessian.copy(), (Vector) src.d0.copy());
    this.isFwd   = src.isFwd;
  }

  public Space getDomain() {
    return isFwd ? getForwardDomain():hessian.getRange();
  }
  public Space getRange () {
    return isFwd ? hessian.getRange():getForwardDomain();
  }
  Space        getForwardDomain()  {
    return ((ProductSpace) hessian.getDomain()).getSubSpace(0);
  }

  public Vector       image(Vector d) {
    return hessian.imageBinary(d0, d);
  }
  public void         image(Vector d, Vector r) {
    hessian.imageBinary(d0, d, r);
  }
  public void      addImage(Vector d, Vector r) {
    hessian.addImageBinary(d0, d, r);
  }
  public void      residual(Vector d, Vector b, Vector r) {
    hessian.residualBinary(d0, d, b, r);
  }

  void createAdjoint() {
    adj = new RHessSubOp(this.hessian, this.d0);
    adj.isFwd = !this.isFwd;
    adj.adj   = this;
  }

  public hasAdjoint     adjoint() {
    if (adj == null) createAdjoint();
    return adj;
  }
  public LinearOperator getGradientAt  (Vector x) { return adjoint(); }
  public hasAdjoint     getDerivativeAt(Vector x) { return this;      }

  public JamObject copy() { return new RHessSubOp(this); }
  public boolean equals(Object obj2) {
    if (this == obj2) return true;
    if (! (obj2 instanceof RHessSubOp)) return false;
    System.err.println("Equals not well implemented for RHessSubOp\n");
    return false;
  }
}

/** the partial adjoint is self-adjoint since the Hessian is symmetric */
final class RHessAdj implements hasAdjoint {
  RosenbrookHess hessian;
  Vector     r0;

  public RHessAdj(RosenbrookHess hessian, Vector rKnown) {
    this.hessian = hessian;
    r0 = rKnown;
  }

  public RHessAdj(RHessAdj src) {
    this((RosenbrookHess) src.hessian.copy(), (Vector) src.r0.copy());
  }

  public Space getRange () { return getDomain(); }
  public Space getDomain() {
    return ((ProductSpace) hessian.getDomain()).getSubSpace(0);
  }

  public Vector       image(Vector d) {
    Vector r = getRange().newMember();
    hessian.apply(false, d, r, r0, RosenbrookHess.kDomain2);
    return r;
  }
  public void         image(Vector d, Vector r) {
    hessian.apply(false, d, r, r0, RosenbrookHess.kDomain2);
  }
  public void      addImage(Vector d, Vector r) {
    hessian.apply(true , d, r, r0, RosenbrookHess.kDomain2);
  }
  public void      residual(Vector d, Vector b, Vector r) {
    r.neg(b);
    hessian.apply(true , d, r, r0, RosenbrookHess.kDomain2);
  }

  public hasAdjoint     adjoint()                 { return this; }
  public LinearOperator getGradientAt  (Vector x) { return this; }
  public hasAdjoint     getDerivativeAt(Vector x) { return this; }

  public JamObject copy() { return new RHessAdj(this); }
  public boolean equals(Object obj2) {
    if (this == obj2) return true;
    if (! (obj2 instanceof RHessAdj)) return false;
    System.err.println("Equals not well implemented for RHessAdj\n");
    return false;
  }
}
