package juice.solver;

import  jam.solver.*;
import  jam.vector.*;
import  jam.operator.*;

/**
 * Conjugate Gradient Solver.<BR>
 * <BR><!--#include file="./source.inc"--><BR>
 * By decomposing the calculations into several routines, <TT>CGPrecondSolver</TT>
 * is made to be subclassed, so that slight modifications to the standard
 * behavior can be made.
 *
 * @author    Joel Schroeder
 * @version   16 Mar 1997 Matt Schwab: added Iterator mechanism
 */
public class CGPrecondSolver extends IterativeSlv implements LinearSolver {

  protected Operator M;          //preconditioner 
  protected Vector   res;        //residual 
  protected float    rtrNew;     //norm2 of the residual
  protected int      reset = 50; //after each reset iterations the exact 

  public CGPrecondSolver(Operator preconditioner) {
    this(new IteratorSimple(), preconditioner); 
  }

  public CGPrecondSolver(Iterator iterator, Operator preconditioner) {
    this.M        = preconditioner;
    this.iterator = iterator;
    this.iterator.setSolver(this); 
  }

  /**
   * after each reset-th iteration I compute the exact residual to 
   * remove accumulated floating point error (default is 50).
   */
  public void setReset(int reset) {
    this.reset = reset;
  }

  public Vector getResidual() {
    return res;
  }
  public float  getNormResidual() {
    return rtrNew;
  }

  public void solve(Operator A, Vector b, Vector x ) { 
    if (! (A instanceof LinearOperator)) {
      System.err.println("Error in CGPrecondSolver.solve(): " +
			 "operator A must be a LinearOperator");
    }

    this.A = A;
    this.b = b;
    this.x = x;

    res = A.getRange().newMember();
    A.residual(x,b,res);
    res.neg();
    Vector p = M.image(res);
    Vector v = A.getRange().newMember();
    Vector s = M.getRange().newMember();

    rtrNew      = res.dot(p);

    iterator.start();

    while(iterator.hasMoreIterations()) {
      doOneStep(s, v, p);
      iterator.nextIteration(); 
      iter++;
    }
    iterator.finish();
  }

  /*
  protected Vector getInitialResidual(Operator A, Vector b, Vector x) {
    // res = b - Ax
    Vector res = A.getRange().newMember();
    A.residual(x, b, res);
    res.neg();
    return res;
  }
  */

  //it is probably better style to allocate s,v, and p when needed.
  //am i thinking again of efficiency?
  protected void doOneStep(Vector s, Vector v, Vector p) {
    A.image(p, v);

    // compute step length alpha
    float ptv   = p.dot( v);
    float alpha = rtrNew/ptv;
    if (ptv == 0f) System.err.println("Error in CGPrecondSolver: " +
				      "divide by ptv (=0).");
	
    // update x and r
	
    x.addScale(  alpha, p);
    if (((iter+1) % reset) == 0) {
      A.residual(x,b,res);
      res.neg();
    }
    else {
      res.addScale( -alpha, v);
    }
    s = M.image(res);

    // compute beta and the new search direction p
    
    float rtrOld = rtrNew;
    rtrNew       = res.dot(s);
    float beta = rtrNew/rtrOld;
    if (rtrOld == 0f) System.err.println("Error in CGPrecondSolver: " +
				      "divide by rtrOld (=0).");
    p.addScale( beta, s, p);
  }

  
  // The following is to be added to the beginning of solve()
  /*
    if (!( A.getDomain()
      .equals(
      A.getRange ()) )) {
      System.err.println( "Error in conjugate gradient solver:\n" +
      "      Domain of A not equal to Range of A." );
      return;
    }
      if (!( A.getRange ().hasMember( b ) )) {
      System.err.println( "Error in conjugate gradient solver:" +
      "      b not in Range of A.");
      return;
    }
    if (!( A.getDomain().hasMember( x ) )) {
      System.err.println(  "Error in conjugate gradient solver:\n"
      + "      x not in Domain of A.\n" );
      return;
    }
    */


}
