package juice.solver;

import  jam.solver.*;
import  jam.vector.*;
import  jam.operator.*;

/**
 * Conjugate Gradient Solver.<BR>
 * <BR><!--#include file="./source.inc"--><BR>
 * By decomposing the calculations into several routines, <TT>CGSolver</TT>
 * is made to be subclassed, so that slight modifications to the standard
 * behavior can be made.
 *
 * @author    Joel Schroeder
 * @version   16 Mar 1997 Matt Schwab: added Iterator mechanism
 */
public class CGSolver extends IterativeSlv implements LinearSolver {

  protected Vector res;        //residual 
  protected float  rtrNew;     //norm2 of the residual
  protected int    reset = 50; //after each reset iterations the exact 

  public CGSolver() {
    this(new IteratorSimple()); 
  }

  public CGSolver(Iterator iterator) {
    this.iterator = iterator;
    this.iterator.setSolver(this); 
  }

  /**
   * after each reset-th iteration I compute the exact residual to 
   * remove accumulated floating point error (default is 50).
   */
  public void setReset(int reset) {
    this.reset = reset;
  }

  public Vector getResidual() {
    return res;
  }
  public float  getNormResidual() {
    return rtrNew;
  }

  public void solve(Operator A, Vector b, Vector x ) { 
    if (! (A instanceof LinearOperator)) {
      System.err.println("Error in CGSolver.solve(): " +
			 "operator A must be a LinearOperator");
    }

    this.A = A;
    this.b = b;
    this.x = x;

    res = A.getRange().newMember();
    A.residual(x,b,res);
    res.neg();
    Vector p = (Vector) res.copy();
    Vector v = A.getRange ().newMember();

    rtrNew      = res.norm2();

    iterator.start();

    while(iterator.hasMoreIterations()) {
      doOneStep(v, p);
      iterator.nextIteration(); 
      iter++;
    }
    iterator.finish();
  }

  protected void doOneStep(Vector v, Vector p) {
    A.image(p, v);

    // compute step length alpha
    float ptv   = p.dot( v);
    float alpha = rtrNew/ptv;
    if (ptv == 0f) System.err.println("Error in CGSolver: " +
				      "divide by ptv (=0) " +
				      "at iter=" + iter + ".");

    // update x and r
	
    x.addScale(  alpha, p);
    if (((iter+1) % reset) == 0) {
      A.residual(x,b,res);
      res.neg();
    }
    else {
      res.addScale( -alpha, v);
    }
	
    // compute beta and the new search direction p
    
    float rtrOld = rtrNew;
    rtrNew       = res.norm2();
    float beta = rtrNew/rtrOld;
    if (rtrOld == 0f) System.err.println("Error in CGSolver: " +
				      "divide by rtrOld (=0).");
    p.addScale( beta, res, p);
  }

  
  // The following is to be added to the beginning of solve()
  /*
    if (!( A.getDomain()
      .equals(
      A.getRange ()) )) {
      System.err.println( "Error in conjugate gradient solver:\n" +
      "      Domain of A not equal to Range of A." );
      return;
    }
      if (!( A.getRange ().hasMember( b ) )) {
      System.err.println( "Error in conjugate gradient solver:" +
      "      b not in Range of A.");
      return;
    }
    if (!( A.getDomain().hasMember( x ) )) {
      System.err.println(  "Error in conjugate gradient solver:\n"
      + "      x not in Domain of A.\n" );
      return;
    }
    */
}
