package juice.solver;

import  jam.solver.*;
import  jam.vector.*;
import  jam.operator.*;
import  java.util.Random;
import  java.lang.Math;

/**
 * Simulated Annealing Solver.<BR>
 * <BR><!--#include file="./source.inc"--><BR
 *
 * @author    Matt Schwab
 * @version   22 Mar 1997
 */

public class SimAnnealSolver extends IterativeSlv {
  float temp;
  ProxyFactory proxyFactory;
  SimAnnealMaster thermo; 

  public SimAnnealSolver(ProxyFactory factory) {
    this(new IteratorSimple(), factory);
  }
  public SimAnnealSolver(Iterator iterator, ProxyFactory factory) {
    this.iterator     = iterator;
    this.iterator.setSolver(this);
    this.proxyFactory = factory; 
    this.thermo       = new StandardMaster(100f,100,100); 
  }

  public void solve(Operator A, Vector b, Vector x) {
    this.A = A;
    this.b = b;
    this.x = x;

    thermo.setTemperature(proxyFactory.getEstimateOfNormResidual()); 
System.out.println("starting temperature : " + thermo.getTemperature());
    iterator.start();

    Proxy proxy;
    while(iterator.hasMoreIterations() && thermo.hasMoreIterations()){

      proxy = proxyFactory.createProxy(A,x,b, thermo.getTemperature());
      thermo.setApproximation(proxy);  

      iter++;
      iterator.nextIteration();
    }
System.out.println("iterator: " + iterator.hasMoreIterations());
System.out.println("thermo  : " +   thermo.hasMoreIterations());
    iterator.finish();
  }

}

interface SimAnnealMaster {
  /**
   * sets a starting Temperature. The temperature is later modified 
   * as we approximate the solution.
   */
  public void setTemperature(float temperature); 
  /** 
   * returns the current temperature of the annealing process;
   */
  public float getTemperature();
  /**
   * returns false when we have cooled sufficiently.
   */
  public boolean hasMoreIterations();
  /**
   * returns the Vector that is the best approximation at this state.
   * The method invokes a simulation generator to inquire about 
   * the energy difference of an alternative approximation. 
   * Based on this energy difference, the method decides if it accepts 
   * the alternative approximation. 
   * Additionally, the temperature of the AnnealingMaster is adjusted 
   * if necessary. 
   */
  public void setApproximation(Proxy proxy); 
}

class StandardMaster implements SimAnnealMaster {
  private float   temperature;
  private boolean hasMoreIterations = true;
  private int temperatureSteps, stepsAtTemperature; 
  private int counterTemperature = 0;
  private int counterSteps = 0;
  private Random  random; 

  StandardMaster(float temperature, 
		 int temperatureSteps, 
		 int stepsAtTemperature) {
    this.temperature        = temperature;
    this.temperatureSteps   = temperatureSteps;
    this.stepsAtTemperature = stepsAtTemperature;
    this.random             = new Random();
  }

  public void setTemperature(float temperature) {
    this.temperature = temperature;
  }

  //annealing schedule
  public void setTemperature() {
    this.temperature *= 0.5f; 
System.out.println("new temperature: " + this.temperature);
    counterTemperature++; 
    counterSteps = 0;
  }

  public float getTemperature() {
    return temperature;
  }

  public boolean hasMoreIterations() { return hasMoreIterations; }

  public void setApproximation(Proxy proxy) {
    double denergy = proxy.getNormResidual() - temperature;
    
    // Metropolis scheme: always takes a down hill step
    if (denergy < 0f || Math.exp(-denergy/temperature) > random.nextFloat()) {
      proxy.setApproximation();
    }
    counterSteps++;
    if (counterSteps       >= stepsAtTemperature) setTemperature();
    if (counterTemperature >= temperatureSteps)   hasMoreIterations = false;
//System.out.println("counterSteps: "       + counterSteps      );
//System.out.println("counterTemperature: " + counterTemperature);

  }
}

