package juice.solver;

import  jam.solver.*;
import  jam.vector.*;
import  jam.operator.*;

/**
 * Simplex Solver.<BR>
 * <BR><!--#include file="./source.inc"--><BR
 * Downhill Simplex method due to Nelder and Maed. 
 * This is distinct from Dantzig's Simplex method for linear 
 * programming problems. For reference see Numerical Recipies (Press, 1992).
 *
 * @author    Matt Schwab
 * @version   4 Mar 1997*
 */

public class SimplexSolver extends IterativeSlv {
  protected Vector[] vertexList;
  protected Simplex  simply;

  // constructors
  public SimplexSolver(Vector[] basis) {
    this(new IteratorSimple(), basis); 
  }
  public SimplexSolver(Iterator iterator, Vector[] basis) {
    this.iterator = iterator;
    this.iterator.setSolver(this);
    vertexList = new Vector[basis.length +1];
    for (int i=0; i < basis.length; i++) {
      vertexList[i] = basis[i];
    }
  }

  // std methods (individual method's may have better ways of computing 
  // these methods. These methods may never be executed. In this case 
  // these standard methods are overwritten. 

  /**
   * A copy would be nicer: I have to loop over the entire data set.
   * This would not happen if I had incorporated x into vertexList 
   * instead of copying it. Don't know how to do this though. 
   */
  public Vector getApproximation() { 
    x.zero();
    x.add(simply.getVertex(simply.getLowest()));
    return x;
  }

  // solve 

  public void solve(Operator A, Vector b, Vector x) {
    this.A = A;
    this.b = b;
    this.x = x;

    setVertexList(x);

    simply = new Simplex(vertexList, A, b);

    iterator.start();

    while(iterator.hasMoreIterations() && currentTol() > 1.0e-12F) {
      simply.setVertexExtrema();
      simply.move();
      iter = iter + vertexList.length + 2;

      iterator.nextIteration();
    }

    iterator.finish();

    x = getApproximation();
  }

  private void setVertexList(Vector x) {
    // it would be nice to have a copy constructor!!
    vertexList[vertexList.length-1] = x.getSpace().newMember();
    vertexList[vertexList.length-1].zero();
    vertexList[vertexList.length-1].add(x);
    for (int i=0; i < vertexList.length - 1; i++) {
      vertexList[i].add(x);
    }
  }
  // Simplex way of measuring accuracy: all functional values within bracket  
  private float currentTol() {
    return (2 * Math.abs(simply.getVertexVal(simply.getHighest()) - 
                         simply.getVertexVal(simply.getLowest())) /
                Math.abs(simply.getVertexVal(simply.getHighest()) + 
		         simply.getVertexVal(simply.getLowest())) );
  }
}


/** 
 * this simplex currently requires the user to hand in a set of 
 * vertices that spans a fully dimensional simplex. The programmer
 * needs to supply these vertices currently since our general 
 * framework lacks the notion of a "canonical basis" for any vector.
 * Every vector space has a basis. But we currently do not require 
 * a programmer to supply that basis (nothing in our vector interface 
 * refers to it). Maybe we should. 
 */ 
class Simplex { 
  //private Operator functional;
  private Operator A;
  private Vector   b; 
  private Vector[] vertexList; 
  private Vector   vertexSum ;
  private float[]  vertexVal ;
  private int      highestVertex, nextHighestVertex, lowestVertex ;

  /** 
   * If I were to hand in a Functional, then I could change the 
   * cost function without ever digging into this beast. I want 
   * to change the cost function since I want to experiment with
   * robust solvers. 
   */
  protected Simplex(Vector[] vertexList, Operator A, Vector b) {
    this.vertexList = vertexList; 
    this.A          = A;
    this.b          = b;
    vertexSum = this.A.getDomain().newMember();
    vertexVal = new float[vertexList.length];
    setVertexVal();
    setVertexSum();
    setVertexExtrema();
  }
  protected void setVertexSum() {
    for(int i = 0; i < vertexList.length; i++) {
      vertexSum.add(vertexList[i]);
    }
  } 
  protected void setVertexVal() {
    Vector res = this.A.getRange().newMember(); 
    for(int i = 0; i < vertexList.length; i++) {
      this.A.residual(vertexList[i], this.b, res);   
      vertexVal[i] = res.norm();
    }
  } 
  protected void setVertexExtrema  () {
    lowestVertex = 1;
    if (vertexVal[1] > vertexVal[2]) {
      highestVertex     = 1;
      nextHighestVertex = 2;
    }
    else {
      highestVertex     = 2;
      nextHighestVertex = 1;
    }
    for(int i = 0; i < vertexList.length; i++) {
      if(vertexVal[i] < vertexVal[lowestVertex ]) {
	lowestVertex = i;
      }
      if(vertexVal[i] > vertexVal[highestVertex]) {
	nextHighestVertex = highestVertex;
	highestVertex     = i;
      }
      else if(vertexVal[i] > vertexVal[nextHighestVertex]) {
	if(i != highestVertex) {nextHighestVertex = i;}
      }
    }
  }
  protected void move() {
    float ytry; 
    ytry   = moveVertex(-1F);
    if      (ytry <= vertexVal[lowestVertex]) {
      ytry = moveVertex(2F);
    }
    else if (ytry >= vertexVal[nextHighestVertex]) { 
      float ysave = vertexVal[highestVertex]; 
      ytry = moveVertex(0.5F); 
      if (ytry >= ysave) {
	contract();
	setVertexVal();
	setVertexSum();
	setVertexExtrema();
      }
    }
  }

  protected int    getHighest       () {return highestVertex;} 
  protected int    getNextHighest   () {return nextHighestVertex;}
  protected int    getLowest        () {return lowestVertex;} 
  protected Vector getVertex   (int i) {return vertexList[i];}
  protected float  getVertexVal(int i) {return vertexVal[i];}

  // you better check this twice!
  protected void  contract   () {
    for(int i = 0; i < vertexList.length; i++) {
      vertexList[i].add(vertexList[lowestVertex]);
      vertexList[i].scale(0.5F); 
    }
  } 
  /**
   * I prefer to have separate try and update step 
   */
  private   float moveVertex  (float step) {
    float step1 = (1F - step) / (vertexList.length - 1);
    float step2 = step1 - step;
    Vector vertexTry;
    vertexTry = this.A.getDomain().newMember();
    vertexTry.addScale(step2, vertexList[highestVertex]);
    vertexTry.neg();
    vertexTry.addScale(step1, vertexSum);

    // maybe we can have this as dummy somewhere
    Vector res = this.A.getRange().newMember(); 
    this.A.residual(vertexTry, this.b, res);   
    float vertexTryVal = res.norm();
    if (vertexTryVal  <= vertexVal[highestVertex]) {
      vertexVal [highestVertex] = vertexTryVal;
      vertexSum.sub(vertexList[highestVertex]);
      vertexSum.add(vertexTry);
      vertexList[highestVertex] = vertexTry;
    }
    return vertexTryVal; 
  } 
}



