package juice.solver;

import  jam.solver.*;
import  jam.vector.*;
import  jam.operator.*;

/**
 * Steepest Descent Solver.<BR>
 * <BR><!--#include file="./source.inc"--><BR>
 * By decomposing the calculations into several routines, <TT>CGSolver</TT>
 * is made to be subclassed, so that slight modifications to the standard
 * behavior can be made.
 *
 * @author    Matt Schwab
 * @version   21 Mar 1997 Matt Schwab
 */
public class SteepDescentSlv extends IterativeSlv implements LinearSolver {

  protected Vector res;        //residual 
  protected float  rtr;        //norm2 of the residual
  protected int    reset = 50; //after each reset iterations the exact 
                               //residual is recalculated. Maybe reset 
                               //could go to IterativeSlv.java ... 

  public SteepDescentSlv() {
    this(new IteratorSimple()); 
  }

  public SteepDescentSlv(Iterator iterator) {
    this.iterator = iterator;
    this.iterator.setSolver(this); 
  }

  /**
   * after each reset-th iteration I compute the exact residual to 
   * remove accumulated floating point error (default is 50).
   */
  public void setReset(int reset) {
    this.reset = reset;
  }

  public Vector getResidual() {
    return res;
  }
  public float  getNormResidual() {
    return rtr;
  }

  public void solve(Operator A, Vector b, Vector x ) { 
    if (! (A instanceof LinearOperator)) {
      System.err.println("Error in SteepDescentSolver.solve(): " +
			 "operator A must be a LinearOperator");
    }

    this.A = A;
    this.b = b;
    this.x = x;

    Vector v = A.getRange().newMember();
    res      = A.getRange().newMember();
    A.residual(x,b,res);
    res.neg();

    rtr      = res.norm2();

    iterator.start();

    while(iterator.hasMoreIterations()) {
      doOneStep(v);
      iterator.nextIteration(); 
      iter++;
    }
    iterator.finish();
  }

  protected void doOneStep(Vector v) {
    A.image(res, v);
    
    // compute step length alpha
    
    float rtv   = res.dot(v);
    float alpha = rtr/rtv;
    if (rtv == 0f) System.err.println("Error in SteepDescentSlv: " +
				      "divide by rtv (=0).");
    
    // update x and r
    // every reset-th iteration I recalculate the exact residual to 
    // remove the accumulated floating point error. 
    x.addScale(  alpha, res);
    if (((iter+1) % reset) == 0) {
      A.residual(x,b,res);
      res.neg();
    }
    else {
      res.addScale( -alpha, v);
    }
    rtr         = res.norm2();
  }
}
